# ----------------------------------------------------------------------
# Copyright (c) 2022
#
# See the LICENSE file for details
# see the AUTHORS file for authors
# ----------------------------------------------------------------------

# ---------------------
# Third party libraries
# ---------------------

import enum


class CentralTendency(enum.Enum):
    MEDIAN = "median"
    MODE = "mode"
    MEAN = "mean"

    def __repr__(self):
        return self.value

    def __str__(self):
        return self.value


class Calibration(enum.Enum):
    MANUAL = "MANUAL"
    AUTO = "AUTO"

    def __repr__(self):
        return self.value

    def __str__(self):
        return self.value


SERIAL_PORT_PREFIX = "/dev/ttyUSB"

# TESS-W data

TEST_IP = "192.168.4.1"
TEST_TCP_PORT = 23
TEST_UDP_PORT = 2255
TEST_SERIAL_PORT = 0
TEST_BAUD = 9600

# Timestamp format
TSTAMP_FORMAT = "%Y-%m-%dT%H:%M:%S.%f"

# Condensed timestamp
TSTAMP_SESSION_FMT = "%Y-%m-%dT%H:%M:%S"



# ----------------------------------------------------------------------
# Copyright (c) 2024 Rafael Gonzalez.
#
# See the LICENSE file for details
# ----------------------------------------------------------------------

# --------------------
# System wide imports
# -------------------

import os
import csv
import logging
import datetime

from argparse import ArgumentParser, Namespace

# -------------------
# Third party imports
# -------------------


from sqlalchemy import select
from sqlalchemy.ext.asyncio import AsyncSession as AsyncSessionClass
from sqlalchemy.ext.asyncio import async_sessionmaker

from lica.sqlalchemy import sqa_logging
from lica.sqlalchemy.asyncio.dbase import AsyncSession
from lica.asyncio.cli import execute

# --------------
# local imports
# -------------

from ... import __version__
from ..util import parser as prs
from ...lib.dbase.model import Config, Round, Photometer, Sample, Summary, Batch

# ----------------
# Module constants
# ----------------

DESCRIPTION = "TESS-W Calibration Database data loader tool"

# -----------------------
# Module global variables
# -----------------------

ORPHANED_SESSIONS_IN_ROUNDS = set()
ORPHANED_SESSIONS_IN_SAMPLES = set()

# get the root logger
log = logging.getLogger(__name__.split(".")[-1])

# -------------------
# Auxiliary functions
# -------------------


async def load_batch(path: str, async_session: async_sessionmaker[AsyncSessionClass]) -> None:
    async with async_session() as session:
        async with session.begin():
            log.info("loading batch data from %s", path)
            with open(path, newline="") as f:
                reader = csv.DictReader(f, delimiter=";")
                for row in reader:
                    row["email_sent"] = True if row["email_sent"] == "1" else False
                    row["begin_tstamp"] = (
                        datetime.datetime.strptime(row["begin_tstamp"], "%Y-%m-%dT%H:%M:%S")
                        if row["begin_tstamp"]
                        else None
                    )
                    row["end_tstamp"] = (
                        datetime.datetime.strptime(row["end_tstamp"], "%Y-%m-%dT%H:%M:%S")
                        if row["end_tstamp"]
                        else None
                    )
                    batch = Batch(**row)
                    log.info("%r", batch)
                    session.add(batch)


async def load_config(path: str, async_session: async_sessionmaker[AsyncSessionClass]) -> None:
    async with async_session() as session:
        async with session.begin():
            log.info("loading config from %s", path)
            with open(path, newline="") as f:
                reader = csv.DictReader(f, delimiter=";")
                for row in reader:
                    if row["section"] == "database" and row["prop"] == "version":
                        value = int(row["value"]) + 1
                        row["value"] = f"{value:02d}"
                    config = Config(**row)
                    log.info("%r", config)
                    session.add(config)


async def load_photometer(path: str, async_session: async_sessionmaker[AsyncSessionClass]) -> None:
    async with async_session() as session:
        async with session.begin():
            log.info("loading photometer from %s", path)
            with open(path, newline="") as f:
                reader = csv.DictReader(f, delimiter=";")
                for row in reader:
                    for key in ("sensor", "firmware", "filter", "collector", "comment"):
                        row[key] = None if not row[key] else row[key]
                    row["freq_offset"] = 0.0
                    phot = Photometer(**row)
                    log.info("%r", phot)
                    session.add(phot)


async def _load_summary(path: str, async_session: async_sessionmaker[AsyncSessionClass]) -> None:
    async with async_session() as session:
        async with session.begin():
            log.info("loading summary from %s", path)
            with open(path, newline="") as f:
                reader = csv.DictReader(f, delimiter=";")
                for row in reader:
                    mac = row["mac"]
                    name = row["name"]
                    del row["mac"]
                    del row["name"]
                    row["session"] = datetime.datetime.strptime(row["session"], "%Y-%m-%dT%H:%M:%S")
                    row["upd_flag"] = True if row["upd_flag"] == "1" else False
                    row["calibration"] = None if not row["calibration"] else row["calibration"]
                    for key in ("zero_point", "zp_offset", "prev_zp", "freq", "mag"):
                        row[key] = float(row[key]) if row[key] else None
                    for key in ("zero_point_method", "freq_method", "nrounds", "comment"):
                        row[key] = None if not row[key] else row[key]
                    log.info("[%9s - %s]Processing row. %s", name, mac, row)
                    q = select(Photometer).where(Photometer.mac == mac, Photometer.name == name)
                    phot = (await session.scalars(q)).one()
                    summary = Summary(**row)
                    summary.photometer = phot
                    log.info("%r", summary)
                    session.add(summary)


async def _assign_batches(async_session: async_sessionmaker[AsyncSessionClass]) -> None:
    async with async_session() as session:
        async with session.begin():
            q = select(Batch)
            batches = (await session.scalars(q)).all()
            for batch in batches:
                log.info("Handling batch %s", batch)
                q = select(Summary).where(
                    Summary.session.between(batch.begin_tstamp, batch.end_tstamp)
                )
                summaries = (await session.scalars(q)).all()
                for summary in summaries:
                    log.info("Assigning summary %s to batch %s", summary, batch)
                    summary.batch = batch
                    session.add(summary)


async def load_summary(path: str, async_session: async_sessionmaker[AsyncSessionClass]) -> None:
    await _load_summary(path, async_session)
    await _assign_batches(async_session)


async def assign_batches(async_session: async_sessionmaker[AsyncSessionClass]) -> None:
    await _assign_batches(async_session)


async def load_rounds(path: str, async_session: async_sessionmaker[AsyncSessionClass]) -> None:
    async with async_session() as session:
        async with session.begin():
            log.info("loading rounds from %s", path)
            with open(path, newline="") as f:
                reader = csv.DictReader(f, delimiter=";")
                for row in reader:
                    row["seq"] = row["round"]
                    del row["round"]
                    meas_session = datetime.datetime.strptime(row["session"], "%Y-%m-%dT%H:%M:%S")
                    row["begin_tstamp"] = (
                        datetime.datetime.strptime(row["begin_tstamp"], "%Y-%m-%dT%H:%M:%S.%f")
                        if row["begin_tstamp"]
                        else None
                    )
                    row["end_tstamp"] = (
                        datetime.datetime.strptime(row["end_tstamp"], "%Y-%m-%dT%H:%M:%S.%f")
                        if row["end_tstamp"]
                        else None
                    )
                    for key in ("freq", "stddev", "mag", "zp_fict", "zero_point", "duration"):
                        row[key] = float(row[key]) if row[key] else None
                    q = select(Summary).where(
                        Summary.session == meas_session, Summary.role == row["role"]
                    )
                    summary = (await session.scalars(q)).one_or_none()
                    if not summary:
                        log.warn(
                            "No summary for round: session=%(session)s seq=%(seq)s, role=%(role)s,",
                            row,
                        )
                        ORPHANED_SESSIONS_IN_ROUNDS.add(meas_session)
                        continue
                    del row["session"]
                    round_ = Round(**row)
                    round_.summary = summary
                    log.info("%r", round_)
                    session.add(round_)
    log.warn("###########################")
    log.warn("ORPHANED SESSIONS IN ROUNDS")
    log.warn("###########################")
    for s in sorted(ORPHANED_SESSIONS_IN_ROUNDS):
        log.warn(s)


async def load_samples(path, async_session: async_sessionmaker[AsyncSessionClass]) -> None:
    async with async_session() as session:
        async with session.begin():
            log.info("loading samples from %s", path)
            with open(path, newline="") as f:
                reader = csv.DictReader(f, delimiter=";")
                for row in reader:
                    meas_session = datetime.datetime.strptime(row["session"], "%Y-%m-%dT%H:%M:%S")
                    del row["session"]
                    row["tstamp"] = datetime.datetime.strptime(
                        row["tstamp"], "%Y-%m-%dT%H:%M:%S.%f"
                    )
                    row["temp_box"] = float(row["temp_box"]) if row["temp_box"] else None
                    row["freq"] = float(row["freq"])
                    row["seq"] = int(row["seq"]) if row["seq"] else None
                    sample = Sample(**row)
                    q = (
                        select(Round)
                        .join(Summary)
                        .where(Summary.session == meas_session, Summary.role == row["role"])
                    )
                    rounds_per_summary = (await session.scalars(q)).all()
                    if not rounds_per_summary:
                        ORPHANED_SESSIONS_IN_SAMPLES.add(meas_session)
                        log.warn("Can't find session %s for this sample %s", meas_session, sample)
                        continue
                    q = select(Summary).where(
                        Summary.session == meas_session, Summary.role == row["role"]
                    )
                    sample.summary = (await session.scalars(q)).one()
                    for r in rounds_per_summary:
                        if r.begin_tstamp <= sample.tstamp <= r.end_tstamp:
                            sample.rounds.append(r)  # no need to do r.append(sample) !!!
                    log.info("%r", sample)
                    session.add(sample)
    log.warn("============================")
    log.warn("ORPHANED SESSIONS IN SAMPLES")
    log.warn("============================")
    for s in sorted(ORPHANED_SESSIONS_IN_SAMPLES):
        log.warn(s)


# --------------
# main functions
# --------------

TABLE = {
    "config": load_config,
    "batch": load_batch,
    "photometer": load_photometer,
    "summary": load_summary,
    "assign": assign_batches,
    "rounds": load_rounds,
    "samples": load_samples,
}


async def loader(args) -> None:
    if args.command not in ("all", "nosamples", "norounds", "assign"):
        func = TABLE[args.command]
        path = os.path.join(args.input_dir, args.command + ".csv")
        await func(path, AsyncSession)
        if args.command == "all":
            assert ORPHANED_SESSIONS_IN_ROUNDS == ORPHANED_SESSIONS_IN_SAMPLES, (
                f"Difference is {ORPHANED_SESSIONS_IN_ROUNDS - ORPHANED_SESSIONS_IN_SAMPLES}"
            )
    elif args.command == "norounds":
        for name in ("config", "batch", "photometer", "summary"):
            path = os.path.join(args.input_dir, name + ".csv")
            func = TABLE[name]
            await func(path, AsyncSession)
    elif args.command == "nosamples":
        for name in ("config", "batch", "photometer", "summary", "rounds"):
            path = os.path.join(args.input_dir, name + ".csv")
            func = TABLE[name]
            await func(path, AsyncSession)
    elif args.command == "assign":
        func = TABLE[name]
        await func(AsyncSession)
    else:
        for name in ("config", "batch", "photometer", "summary", "rounds", "samples"):
            path = os.path.join(args.input_dir, name + ".csv")
            func = TABLE[name]
            await func(path, AsyncSession)


def add_args(parser: ArgumentParser) -> None:
    subparser = parser.add_subparsers(dest="command")
    subparser.add_parser("config", parents=[prs.idir()], help="Load config CSV")
    subparser.add_parser("batch", parents=[prs.idir()], help="Load batch CSV")
    subparser.add_parser("photometer", parents=[prs.idir()], help="Load photometer CSV")
    subparser.add_parser("summary", parents=[prs.idir()], help="Load summary CSV")
    subparser.add_parser("assign", parents=[], help="Assign summaries to batches")
    subparser.add_parser("rounds", parents=[prs.idir()], help="Load rounds CSV")
    subparser.add_parser("samples", parents=[prs.idir()], help="Load samples CSV")
    subparser.add_parser("nosamples", parents=[prs.idir()], help="Load all CSVs except samples")
    subparser.add_parser("all", parents=[prs.idir()], help="Load all CSVs")


async def cli_main(args: Namespace) -> None:
    sqa_logging(args)
    await loader(args)


def main():
    """The main entry point specified by pyproject.toml"""
    execute(
        main_func=cli_main,
        add_args_func=add_args,
        name=__name__,
        version=__version__,
        description=DESCRIPTION,
    )


if __name__ == "__main__":
    main()



# ----------------------------------------------------------------------
# Copyright (c) 2024 Rafael Gonzalez.
#
# See the LICENSE file for details
# ----------------------------------------------------------------------

# --------------------
# System wide imports
# -------------------

import logging
from argparse import ArgumentParser, Namespace

# ------------------
# SQLAlchemy imports
# -------------------

from lica.asyncio.cli import execute
from lica.sqlalchemy import sqa_logging
from lica.sqlalchemy.asyncio.dbase import engine
from lica.sqlalchemy.asyncio.model import Model
# --------------
# local imports
# -------------

from ... import __version__

# We must pull one model to make it work
from ...lib.dbase.model import Config  # noqa: F401

# ----------------
# Module constants
# ----------------

DESCRIPTION = "TESS-W Calibration Database initial schema generation tool"

# -----------------------
# Module global variables
# -----------------------

# get the module logger
log = logging.getLogger(__name__.split(".")[-1])

# -------------------
# Auxiliary functions
# -------------------


async def schema() -> None:
    async with engine.begin() as conn:
        await conn.run_sync(Model.metadata.drop_all)
        await conn.run_sync(Model.metadata.create_all)
    await engine.dispose()


async def cli_main(args: Namespace) -> None:
    sqa_logging(args)
    await schema()


def add_args(parser: ArgumentParser) -> None:
    pass


def main():
    """The main entry point specified by pyproject.toml"""
    execute(
        main_func=cli_main,
        add_args_func=add_args,
        name=__name__,
        version=__version__,
        description=DESCRIPTION,
    )


if __name__ == "__main__":
    main()



# ----------------------------------------------------------------------
# Copyright (c) 2024 Rafael Gonzalez.
#
# See the LICENSE file for details
# ----------------------------------------------------------------------

# --------------------
# System wide imports
# -------------------


import math
import logging
from argparse import ArgumentParser, Namespace
from datetime import datetime
import statistics

# -------------------
# Third party imports
# -------------------

from lica.validators import vdate


from sqlalchemy import update
from sqlalchemy.ext.asyncio import AsyncSession as AsyncSessionClass
from sqlalchemy.ext.asyncio import async_sessionmaker

from lica.sqlalchemy import sqa_logging
from lica.sqlalchemy.asyncio.dbase import AsyncSession
from lica.asyncio.cli import execute


# --------------
# local imports
# -------------

from ...lib.dbase.model import Round
from ... import __version__
from ...lib import CentralTendency

# ----------------
# Module constants
# ----------------

DESCRIPTION = "TESS-W Calibration Database fix stuff tool"

# -----------------------
# Module global variables
# -----------------------

log = logging.getLogger(__name__.split(".")[-1])

# -------------------
# Auxiliary functions
# -------------------


def central(method: str):
    assert method in CentralTendency, f"Statistics central method {method} is not supported"
    f = statistics.mode
    if method == CentralTendency.MEAN.value:
        f = statistics.mean
    elif method == CentralTendency.MEDIAN.value:
        f = statistics.median
    return f


def rounds(conn, session, role):
    params = {"session": session, "role": role}
    cursor = conn.cursor()
    cursor.execute(
        """
        SELECT DISTINCT begin_tstamp, end_tstamp, round, freq, central, stddev
        FROM rounds_t
        WHERE session = :session AND role = :role
        AND begin_tstamp IS NOT NULL AND end_tstamp IS NOT NULL
    """,
        params,
    )
    return cursor


def samples(conn, session, begin_tstamp, end_tstamp, role):
    params = {
        "session": session,
        "role": role,
        "begin_tstamp": begin_tstamp,
        "end_tstamp": end_tstamp,
    }
    cursor = conn.cursor()
    cursor.execute(
        """
        SELECT freq FROM samples_t
        WHERE session = :session AND role = :role
        AND tstamp BETWEEN :begin_tstamp AND :end_tstamp
        ORDER BY tstamp
    """,
        params,
    )
    return list(result[0] for result in cursor)


def fix_stddev(conn, new_stddev, old_stddev, name, mac, session, role, seq):
    params = {
        "session": session,
        "role": role,
        "seq": seq,
        "new_stddev": new_stddev,
        "old_stddev": old_stddev,
    }
    with conn:
        cursor = conn.cursor()
        cursor.execute(
            """
            UPDATE rounds_t
            SET stddev = :new_stddev       
            WHERE session = :session AND role = :role AND round = :seq
            AND stddev = :old_stddev
        """,
            params,
        )
        log.info(
            "[%s] [%s] [%s] Round #%d update old %f => new %f",
            name,
            mac,
            session,
            seq,
            old_stddev,
            new_stddev,
        )


def compare_and_fix_stddev(
    conn, dry_run, name, mac, session, role, seq, freq, freqs, freq_method, stddev
):
    central_func = central(freq_method)
    computed_freq = central_func(freqs)
    computed_stddev = statistics.stdev(freqs, computed_freq)
    if not math.fabs(computed_stddev - stddev) < 0.005:
        log.warn(
            "[%s] [%s] [%s] Round #%d computed \u03c3(%s %f)=%f, != stored \u03c3(%f)=%f",
            name,
            mac,
            session,
            seq,
            freq_method,
            computed_freq,
            computed_stddev,
            freq,
            stddev,
        )
        freq2 = statistics.mean(freqs)
        computed2_stddev = statistics.stdev(freqs, freq2)
        if not math.fabs(computed2_stddev - stddev) < 0.005:
            log.error(
                "[%s] [%s] [%s] Round #%d Computed \u03c3(%s %f)=%f != stored\u03c3(%f) %f",
                name,
                mac,
                session,
                seq,
                "mean",
                freq2,
                computed2_stddev,
                freq,
                stddev,
            )
        elif not dry_run:
            fix_stddev(conn, computed_stddev, stddev, name, mac, session, role, seq)


def fix_rounds_stddev(conn, dry_run, name, mac, session, role) -> None:
    for begin_tstamp, end_tstsamp, seq, freq, central, stddev in rounds(conn, session, role):
        freqs = samples(conn, session, begin_tstamp, end_tstsamp, role)
        compare_and_fix_stddev(
            conn, dry_run, name, mac, session, role, seq, freq, freqs, central, stddev
        )


def sessions(conn, session=None):
    cursor = conn.cursor()
    if session:
        params = {"session": session}
        sql = "SELECT name, mac, session, role FROM summary_t WHERE session = :session ORDER BY session ASC"
    else:
        params = {}
        sql = "SELECT name, mac, session, role FROM summary_t ORDER BY session ASC"
    cursor.execute(sql, params)
    return cursor


async def fix_fict_zp(
    async_session: async_sessionmaker[AsyncSessionClass],
    meas_session: datetime,
    dry_run: bool,
    default_zp=20.50,
) -> None:
    log.info("Fixing Ficticious ZP to %s", default_zp)
    stmt = update(Round).where(Round.zp_fict != default_zp).values(zp_fict=default_zp)
    if not dry_run:
        async with async_session() as session:
            async with session.begin():
                await session.execute(stmt)


# --------------
# main functions
# --------------


def add_args(parser: ArgumentParser):
    subparser = parser.add_subparsers(dest="command")
    parser_rounds = subparser.add_parser("rounds", help="Fix rounds stuff")
    roex = parser_rounds.add_mutually_exclusive_group(required=True)
    roex.add_argument("-t", "--stddev", action="store_true", help="Fix rounds stddev")
    roex.add_argument("-z", "--zp-fict", action="store_true", help="Fix Ficticious ZP to 20.50")
    parser_rounds.add_argument("-s", "--session", type=vdate, default=None, help="Session date")
    parser_rounds.add_argument(
        "-d", "--dry-run", action="store_true", help="Do not update database"
    )


async def fix(args: Namespace) -> None:
    if args.stddev:
        await fix_rounds_stddev(AsyncSession, args.session)
    elif args.zp_fict:
        await fix_fict_zp(AsyncSession, args.session, args.dry_run)


async def cli_main(args: Namespace) -> None:
    sqa_logging(args)
    await fix(args)


def main():
    """The main entry point specified by pyproject.toml"""
    execute(
        main_func=cli_main,
        add_args_func=add_args,
        name=__name__,
        version=__version__,
        description=DESCRIPTION,
    )


if __name__ == "__main__":
    main()



# ----------------------------------------------------------------------
# Copyright (c) 2024 Rafael Gonzalez.
#
# See the LICENSE file for details
# ----------------------------------------------------------------------

# --------------------
# System wide imports
# -------------------

import os

import csv
import logging

from argparse import ArgumentParser, Namespace
from typing import Sequence, Iterable
from sqlite3 import Connection

# -------------------
# Third party imports
# -------------------

from lica.cli import execute
from lica.sqlite import open_database

# --------------
# local imports
# -------------

from ... import __version__
from ..util import parser as prs

# ----------------
# Module constants
# ----------------

DESCRIPTION = "TESS-W Calibration Database data extraction tool"

CONFIG_H = ("section", "prop", "value")
BATCH_H = ("begin_tstamp", "end_tstamp", "calibrations", "email_sent", "comment")
PHOTOMETER_H = ("name", "mac", "sensor", "model", "firmware", "filter", "plug", "box", "collector", "comment")
SUMMARY_H = (
    "name",
    "mac",
    "session",
    "role",
    "calibration",
    "calversion",
    "author",
    "nrounds",
    "zp_offset",
    "upd_flag",
    "prev_zp",
    "zero_point",
    "zero_point_method",
    "freq",
    "freq_method",
    "mag",
    "comment",
)
ROUNDS_H = (
    "session",
    "round",
    "role",
    "begin_tstamp",
    "end_tstamp",
    "central",
    "freq",
    "stddev",
    "mag",
    "zp_fict",
    "zero_point",
    "nsamples",
    "duration",
)
SAMPLES_H = ("session", "tstamp", "role", "seq", "freq", "temp_box")

# -----------------------
# Module global variables
# -----------------------

# get the root logger
log = logging.getLogger(__name__.split(".")[-1])

# -------------------
# Auxiliary functions
# -------------------


def write_csv(path: str, header: Sequence[str], iterable: Iterable[str], delimiter: str = ";"):
    with open(path, "w", newline="") as csvfile:
        writer = csv.writer(csvfile, delimiter=delimiter)
        writer.writerow(header)
        for row in iterable:
            writer.writerow(row)
    log.info("Written to %s", path)


def _extract_batch(path: str, conn: Connection) -> None:
    log.info("Extracting from batch_t table.")
    cursor = conn.cursor()
    cursor.execute("""
        SELECT begin_tstamp, end_tstamp, calibrations, email_sent, comment 
        FROM batch_t 
        ORDER BY begin_tstamp
    """)
    write_csv(path, BATCH_H, cursor)


def _extract_config(path: str, conn: Connection) -> None:
    log.info("Extracting from config_t table.")
    cursor = conn.cursor()
    cursor.execute("""
        SELECT section, property AS prop, value
        FROM config_t 
        ORDER BY section, prop
    """)
    write_csv(path, CONFIG_H, cursor)


def _extract_photometer(path: str, conn: Connection) -> None:
    log.info("Extracting from summary_t table for photometer data.")
    cursor = conn.cursor()
    cursor.execute("""
        SELECT DISTINCT name,mac,sensor,model,firmware,filter,plug,box,collector,comment
        FROM summary_t
        WHERE comment IS NULL
        UNION
        SELECT DISTINCT name,mac,sensor,model,firmware,filter,plug,box,collector,SUBSTR(comment,7)
        FROM summary_t
        WHERE comment LIKE 'Phot:%'
        UNION
        SELECT DISTINCT name,mac,sensor,model,firmware,filter,plug,box,collector,SUBSTR(comment,11)
        FROM summary_t
        WHERE comment LIKE 'PhotSumm:%'
        UNION
        SELECT DISTINCT name,mac,sensor,model,firmware,filter,plug,box,collector,NULL
        FROM summary_t
        WHERE comment LIKE 'Summ:%'
        ORDER BY name,mac
    """)
    write_csv(path, PHOTOMETER_H, cursor)


def _extract_summary(path: str, conn: Connection) -> None:
    log.info("Extracting from summary_t table for summary calibration data.")
    cursor = conn.cursor()
    cursor.execute("""
        SELECT DISTINCT name,mac,session,role,calibration,calversion,author,nrounds,offset AS zp_offset,
            upd_flag,prev_zp,zero_point,zero_point_method,freq,freq_method,mag,comment
        FROM summary_t
        WHERE comment IS NULL
        UNION
        SELECT DISTINCT name,mac,session,role,calibration,calversion,author,nrounds,offset AS zp_offset,
            upd_flag,prev_zp,zero_point,zero_point_method,freq,freq_method,mag, SUBSTR(comment,7)
        FROM summary_t
        WHERE comment LIKE 'Summ:%'
        UNION
        SELECT DISTINCT name,mac,session,role,calibration,calversion,author,nrounds,offset AS zp_offset,
            upd_flag,prev_zp,zero_point,zero_point_method,freq,freq_method,mag,SUBSTR(comment,11)
        FROM summary_t
        WHERE comment LIKE 'PhotSumm:%'
        UNION
        SELECT DISTINCT name,mac,session,role,calibration,calversion,author,nrounds,offset AS zp_offset,
            upd_flag,prev_zp,zero_point,zero_point_method,freq,freq_method,mag,NULL
        FROM summary_t
        WHERE comment LIKE 'Phot:%'
        ORDER BY name,mac

    """)
    write_csv(path, SUMMARY_H, cursor)


def _extract_rounds(path: str, conn: Connection) -> None:
    log.info("Extracting from rounds_t table.")
    cursor = conn.cursor()
    cursor.execute("""
        SELECT DISTINCT session,round,role,begin_tstamp,end_tstamp,central,freq,stddev,mag,zp_fict,zero_point,nsamples,duration
        FROM rounds_t
        ORDER BY session, round, role
    """)
    write_csv(path, ROUNDS_H, cursor)


def _extract_samples(path: str, conn: Connection) -> None:
    log.info("Extracting from samples_t table.")
    cursor = conn.cursor()
    cursor.execute("""
        SELECT session,tstamp,role,seq,freq,temp_box 
        FROM samples_t 
        ORDER BY session, tstamp, role 
    """)
    write_csv(path, SAMPLES_H, cursor)


# --------------
# main functions
# --------------


def add_args(parser: ArgumentParser) -> None:
    subparser = parser.add_subparsers(dest="command")
    subparser.add_parser("config", parents=[prs.odir()], help="Extract config CSV")
    subparser.add_parser("batch", parents=[prs.odir()], help="Extract batch CSV")
    subparser.add_parser("photometer", parents=[prs.odir()], help="Extract photometer CSV")
    subparser.add_parser("summary", parents=[prs.odir()], help="Extract summary CSV")
    subparser.add_parser("rounds", parents=[prs.odir()], help="Extract rounds CSV")
    subparser.add_parser("samples", parents=[prs.odir()], help="Extract samples CSV")
    subparser.add_parser("all", parents=[prs.odir()], help="Extract all CSVs")


TABLE = {
    "config": _extract_config,
    "batch": _extract_batch,
    "photometer": _extract_photometer,
    "summary": _extract_summary,
    "rounds": _extract_rounds,
    "samples": _extract_samples,
}


def cli_main(args: Namespace) -> None:
    connection, _ = open_database(env_var="SOURCE_DATABASE")
    if args.command not in ("all",):
        func = TABLE[args.command]
        path = os.path.join(args.output_dir, args.command + ".csv")
        func(path, connection)
        log.info("done.")
    else:
        for name in ("config", "batch", "photometer", "summary", "rounds", "samples"):
            path = os.path.join(args.output_dir, name + ".csv")
            func = TABLE[name]
            func(path, connection)
    connection.close()


def main():
    """The main entry point specified by pyproject.toml"""
    execute(
        main_func=cli_main,
        add_args_func=add_args,
        name=__name__,
        version=__version__,
        description=DESCRIPTION,
    )


if __name__ == "__main__":
    main()



# ----------------------------------------------------------------------
# Copyright (c) 2024 Rafael Gonzalez.
#
# See the LICENSE file for details
# ----------------------------------------------------------------------


# --------------------
# System wide imports
# -------------------

import sys
import logging

from typing import Optional, List, Set
from datetime import datetime

# =====================
# Third party libraries
# =====================


if sys.version_info[1] < 11:
    from typing_extensions import Self
else:
    from typing import Self


from sqlalchemy import (
    select,
    func,
    Enum,
    Table,
    Column,
    Integer,
    String,
    DateTime,
    ForeignKey,
    UniqueConstraint,
)
from sqlalchemy.orm import Mapped, mapped_column, relationship, aliased

from lica.sqlalchemy.asyncio.model import Model
from lica.sqlalchemy.view import view
from lica.asyncio.photometer import Model as PhotModel, Role, Sensor

from .. import CentralTendency, Calibration

# ================
# Module constants
# ================

# =======================
# Module global variables
# =======================

# get the module logger
log = logging.getLogger(__name__)


def datestr(dt: datetime) -> str:
    return dt.strftime("%Y-%m-%d %H:%M:%S.%f") if dt is not None else None


# =================================
# Data Model, declarative ORM style
# =================================

# ---------------------------------------------
# Additional conveniente types for enumerations
# ---------------------------------------------

RoleType: Enum = Enum(
    Role,
    name="role_type",
    create_constraint=False,
    metadata=Model.metadata,
    validate_strings=True,
    values_callable=lambda x: [e.name.lower() for e in x],
)

PhotModelType: Enum = Enum(
    PhotModel,
    name="model_type",
    create_constraint=False,
    metadata=Model.metadata,
    validate_strings=True,
    values_callable=lambda x: [e.value for e in x],
)


SensorType: Enum = Enum(
    Sensor,
    name="sensor_type",
    create_constraint=False,
    metadata=Model.metadata,
    validate_strings=True,
    values_callable=lambda x: [e.value for e in x],
)

CentralTendencyType: Enum = Enum(
    CentralTendency,
    name="central_type",
    create_constraint=False,
    metadata=Model.metadata,
    validate_strings=True,
    values_callable=lambda x: [e.value for e in x],
)

CalibrationType: Enum = Enum(
    Calibration,
    name="calibration_type",
    create_constraint=False,
    metadata=Model.metadata,
    validate_strings=True,
    values_callable=lambda x: [e.value for e in x],
)

# --------
# Entities
# --------


class Config(Model):
    __tablename__ = "config_t"

    section: Mapped[str] = mapped_column(String(32), primary_key=True)
    prop: Mapped[str] = mapped_column("property", String(255), primary_key=True)
    value: Mapped[str] = mapped_column(String(255))

    def __repr__(self) -> str:
        return f"Config(section={self.section!r}, prop={self.prop!r}, value={self.value!r})"


class Batch(Model):
    __tablename__ = "batch_t"
    
    id: Mapped[int] = mapped_column(primary_key=True)
    begin_tstamp: Mapped[datetime] = mapped_column(DateTime, unique=True)
    end_tstamp: Mapped[Optional[datetime]] = mapped_column(DateTime)
    email_sent: Mapped[Optional[bool]]
    calibrations: Mapped[Optional[int]]
    comment: Mapped[Optional[str]] = mapped_column(String(255))

    # This is not a real column, it s meant for the ORM
    summaries: Mapped[List["Summary"]] = relationship(back_populates="batch")

    def __repr__(self) -> str:
        return f"Batch(begin={datestr(self.begin_tstamp)}, end={datestr(self.end_tstamp)})"


class Photometer(Model):
    __tablename__ = "photometer_t"

    id: Mapped[int] = mapped_column(primary_key=True)
    name: Mapped[str] = mapped_column(String(10))
    mac: Mapped[str] = mapped_column(String(17))
    sensor: Mapped[SensorType] = mapped_column(SensorType, default=Sensor.TSL237)
    freq_offset: Mapped[float]
    model: Mapped[PhotModelType] = mapped_column(PhotModelType)
    firmware: Mapped[Optional[str]] = mapped_column(String(17))
    filter: Mapped[Optional[str]] = mapped_column(String(32), default="UV/IR-740")
    plug: Mapped[Optional[str]] = mapped_column(String(16), default="USB-A")
    box: Mapped[Optional[str]] = mapped_column(String(16), default="FSH714")
    collector: Mapped[Optional[str]] = mapped_column(
        String(16), default="standard"
    )  #  Collector model
    comment: Mapped[Optional[str]] = mapped_column(String(255))  # Photometer level comment

    # This is not a real column, it s meant for the ORM
    calibrations: Mapped[List["Summary"]] = relationship(back_populates="photometer")

    def __repr__(self) -> str:
        return f"Photom(id={self.id!r}, name={self.name!r}, mac={self.mac!r})"

    __table_args__ = (
        UniqueConstraint(name, mac),
        {},
    )


class Summary(Model):
    __tablename__ = "summary_t"

    id: Mapped[int] = mapped_column(primary_key=True)
    phot_id: Mapped[int] = mapped_column(ForeignKey("photometer_t.id"), index=True)
    batch_id: Mapped[Optional[int]] = mapped_column(ForeignKey("batch_t.id"))
    session: Mapped[datetime] = mapped_column(DateTime)  # calibration session identifier
    role: Mapped[RoleType] = mapped_column(RoleType)
    calibration: Mapped[CalibrationType] = mapped_column(CalibrationType, nullable=True)
    calversion: Mapped[Optional[str]] = mapped_column(String(64))  # calibration software version
    author: Mapped[Optional[str]]  # who run the calibration
    nrounds: Mapped[Optional[int]]  # Number of rounds passed
    zp_offset: Mapped[
        Optional[float]
    ]  # Additional offset that was summed to the computed zero_point
    upd_flag: Mapped[Optional[bool]]  # 1 => TESS-W ZP was updated, 0 => TESS-W ZP was not updated
    prev_zp: Mapped[Optional[float]]
    zero_point: Mapped[Optional[float]]  #  calibrated zero point
    zero_point_method: Mapped[CentralTendencyType] = mapped_column(
        CentralTendencyType, nullable=True
    )
    freq: Mapped[Optional[float]]  # final chosen frequency
    freq_method: Mapped[CentralTendencyType] = mapped_column(CentralTendencyType, nullable=True)
    mag: Mapped[Optional[float]]
    comment: Mapped[Optional[str]] = mapped_column(
        String(512)
    )  #  Additional comment for the calibration process

    # These are not a real columns, it is meant for the ORM
    batch: Mapped[Optional["Batch"]] = relationship(back_populates="summaries")
    photometer: Mapped["Photometer"] = relationship(back_populates="calibrations")
    rounds: Mapped[List["Round"]] = relationship(back_populates="summary")
    samples: Mapped[Set["Sample"]] = relationship(back_populates="summary")

    def __repr__(self) -> str:
        return f"Summary(session={datestr(self.session)}, role={self.role!r}, phot_id={self.phot_id!r})"

    __table_args__ = (UniqueConstraint(session, role), {})


# Samples per round
# Due to the sliding window collect process, a sample may belong to several rounds
# This part is not part of the ORM, as it uses the basic Table API
SamplesRounds = Table(
    "samples_rounds_t",
    Model.metadata,
    Column("round_id", ForeignKey("rounds_t.id"), nullable=False, primary_key=True),
    Column("sample_id", ForeignKey("samples_t.id"), nullable=False, primary_key=True),
)


class Round(Model):
    __tablename__ = "rounds_t"

    id: Mapped[int] = mapped_column(primary_key=True)
    summ_id: Mapped[int] = mapped_column(ForeignKey("summary_t.id"), index=True)
    seq: Mapped[int] = mapped_column("round", Integer)  # Round number form 1..NRounds
    role: Mapped[RoleType] = mapped_column(RoleType)
    # session:    Mapped[datetime] = mapped_column(DateTime)
    freq: Mapped[Optional[float]]
    # Either average or median of samples for this frequencies round
    central: Mapped[CentralTendencyType] = mapped_column(CentralTendencyType, nullable=True)
    stddev: Mapped[Optional[float]]  # Standard deviation for frequency central estimate
    mag: Mapped[
        Optional[float]
    ]  # magnitiude corresponding to central frequency and summing ficticious zero point
    zp_fict: Mapped[Optional[float]]  # Ficticious ZP to estimate instrumental magnitudes (=20.50)
    zero_point: Mapped[
        Optional[float]
    ]  # Estimated Zero Point for this round ('test' photometer round only, else NULL)
    nsamples: Mapped[Optional[int]]  # Number of samples for this round
    duration: Mapped[Optional[float]]  # Approximate duration, in seconds
    begin_tstamp: Mapped[Optional[datetime]] = mapped_column(DateTime)
    end_tstamp: Mapped[Optional[datetime]] = mapped_column(DateTime)

    # This is not a real column, it s meant for the ORM
    summary: Mapped["Summary"] = relationship(back_populates="rounds")
    # samples per round. Shoudl match the window size
    # This is not a real column, it s meant for the ORM
    samples: Mapped[List["Sample"]] = relationship(secondary=SamplesRounds, back_populates="rounds")

    def __repr__(self) -> str:
        return f"Round(id={self.id!r}, #{self.seq!r} [{self.nsamples!r}] {self.role!r}, zp={self.zero_point} f={self.freq}, m={self.mag:.2f}@{self.zp_fict} Ts={datestr(self.begin_tstamp)}, Te={datestr(self.end_tstamp)})"

    __table_args__ = (UniqueConstraint(summ_id, seq, role), {})


class Sample(Model):
    __tablename__ = "samples_t"

    id: Mapped[int] = mapped_column(primary_key=True)
    summ_id: Mapped[int] = mapped_column(ForeignKey("summary_t.id"), index=True)
    tstamp: Mapped[datetime] = mapped_column(DateTime)
    role: Mapped[RoleType] = mapped_column(RoleType)
    seq: Mapped[Optional[int]]
    freq: Mapped[float]
    temp_box: Mapped[Optional[float]]

    # rounds per sample (at least 1...)
    # This is not a real column, it s meant for the ORM
    rounds: Mapped[List["Round"]] = relationship(secondary=SamplesRounds, back_populates="samples")

    # This is not a real column, it s meant for the ORM
    summary: Mapped["Summary"] = relationship(back_populates="samples")

    def __repr__(self) -> str:
        return f"Sample(id={self.id!r}, role={self.role!r} freq={self.freq!r},  seq={self.seq!r})"

    def __lt__(self, other: Self) -> bool:
        return self.tstamp < other.tstamp

    def __le__(self, other: Self) -> bool:
        return self.tstamp <= other.tstamp

    def __eq__(self, other: Self) -> bool:
        return self.tstamp == other.tstamp

    def __ne__(self, other: Self) -> bool:
        return self.tstamp != other.tstamp

    def __gt__(self, other: Self) -> bool:
        return self.tstamp > other.tstamp

    def __ge__(self, other: Self) -> bool:
        return self.tstamp >= other.tstamp

    def __hash__(self):
        return hash(self.tstamp)

    __table_args__ = (UniqueConstraint(tstamp, role), {})


# Create the view for barebones SQL statements from console

Ref_t = aliased(Summary)

summary_view = view(
    name="summary_v",
    metadata=Model.metadata,
    selectable=select(
        Summary.id.label("id"),
        Photometer.model.label("model"),
        Photometer.name.label("name"),
        Photometer.mac.label("mac"),
        Photometer.firmware.label("firmware"),
        Photometer.sensor.label("sensor"),
        Summary.session.label("session"),
        Summary.role.label("role"),
        Summary.nrounds.label("nrounds"),
        Summary.upd_flag.label("upd_flag"),
        func.round(Summary.zero_point, 2).label("zero_point"),
        func.round(Summary.zp_offset, 2).label("zp_offset"),
        func.round((Summary.zero_point - Summary.zp_offset), 2).label("raw_zero_point"),
        Summary.calibration.label("calibration"),
        func.round(Summary.prev_zp, 2).label("prev_zp"),
        func.round(Ref_t.mag, 2).label("ref_mag"),
        func.round(Ref_t.freq, 3).label("ref_freq"),
        func.round(Summary.mag, 2).label("test_mag"),
        func.round(Summary.freq, 3).label("test_freq"),
        func.round((Ref_t.mag - Summary.mag), 2).label("mag_diff"),
        Photometer.freq_offset.label("freq_offset"),
        Summary.zero_point_method.label("zero_point_method"),
        Summary.freq_method.label("freq_method"),
        Summary.calversion.label("calversion"),
        Photometer.filter.label("filter"),
        Photometer.plug.label("plug"),
        Photometer.box.label("box"),
        Photometer.collector.label("collector"),
        Summary.author.label("author"),
        Summary.comment.label("comment"),
    )
    .join(Ref_t, Ref_t.session == Summary.session)
    .join(Photometer, Photometer.id == Summary.phot_id)
    .where(Ref_t.role == Role.REF, Summary.role == Role.TEST),
)

# Another view for exporting data
rounds_view = view(
    name="rounds_v",
    metadata=Model.metadata,
    selectable=select(
        Round.id.label("id"),
        Photometer.name.label("name"),
        Photometer.mac.label("mac"),
        Photometer.model.label("model"),
        Summary.session.label("session"),
        Round.__table__.c.round.label("round"),  # Problems with the 'round' attribute name'
        Round.role.label("role"),
        func.round(Round.freq, 3).label("freq"),
        Round.central.label("central"),
        func.round(Round.stddev, 4).label("stddev"), 
        func.round(Round.mag, 3).label("mag"),
        Round.zp_fict.label("zp_fict"),
        Round.zero_point.label("zero_point"),
        Round.nsamples.label("nsamples"),
        func.round(Round.duration, 3).label("duration"),
        Round.begin_tstamp.label("begin_tstamp"),
        Round.end_tstamp.label("end_tstamp"),
        Summary.upd_flag.label("upd_flag"),
        Summary.nrounds.label("nrounds"),
        Summary.freq.label("mean_freq"),
        Summary.freq_method.label("freq_method"),
    )
    .join(Summary, Round.summ_id == Summary.id)
    .join(Photometer, Photometer.id == Summary.phot_id),
)


# Another view for exporting data
samples_view = view(
    name="samples_v",
    metadata=Model.metadata,
    selectable=select(
        Sample.id.label("id"),
        Photometer.name.label("name"),
        Photometer.mac.label("mac"),
        Photometer.model.label("model"),
        Summary.session.label("session"),
        Summary.upd_flag.label("upd_flag"),
        Round.__table__.c.round.label("round"),  # Problems with the 'round' attribute name'
        Sample.role.label("role"),
        Sample.tstamp.label("tstamp"),
        func.round(Sample.freq, 3).label("freq"),
        Sample.temp_box.label("temp_box"),
        Sample.seq.label("seq"),
    )
    .select_from(SamplesRounds)
    .join(Sample,  SamplesRounds.c.sample_id == Sample.id)
    .join(Round,  SamplesRounds.c.round_id == Round.id)
    .join(Summary, Sample.summ_id == Summary.id)
    .join(Photometer, Photometer.id == Summary.phot_id)
)


class SummaryView(Model):
    __table__ = summary_view

    def __repr__(self) -> str:
        return f"SummaryView(name={self.name}, mac={self.mac}, session={datestr(self.session)}, role={self.role!r}, nrounds={self.nrounds!r}, zp={self.zero_point!r}, calib={self.calibration!r}, freq={self.freq!r})"

class RoundView(Model):
    __table__ = rounds_view

    def __repr__(self) -> str:
        return f"RoundsView(name={self.name}, mac={self.mac}, session={datestr(self.session)}, role={self.role!r}, freq={self.freq!r}, method={self.central!r})"

class SampleView(Model):
    __table__ = samples_view

    def __repr__(self) -> str:
        return f"SampleView(name={self.name}, mac={self.mac}, session={datestr(self.session)}, role={self.role!r}, freq={self.freq!r}, sequence={self.sequence!r})"



# ----------------------------------------------------------------------
# Copyright (c) 2024 Rafael Gonzalez.
#
# See the LICENSE file for details
# ----------------------------------------------------------------------

# --------------------
# System wide imports
# -------------------


import math
import logging
from datetime import datetime
import statistics

from argparse import ArgumentParser, Namespace
from typing import List

# -------------------
# Third party imports
# -------------------

from sqlalchemy import select
from sqlalchemy.ext.asyncio import AsyncSession as AsyncSessionClass
from sqlalchemy.ext.asyncio import async_sessionmaker

from lica.sqlalchemy import sqa_logging
from lica.sqlalchemy.asyncio.dbase import AsyncSession
from lica.asyncio.photometer import Role
from lica.validators import vdate
from lica.asyncio.cli import execute

# --------------
# local imports
# -------------

from ... import __version__
from ...lib.dbase.model import Round, Photometer, Sample, Summary
from ...lib import CentralTendency

# ----------------
# Module constants
# ----------------

DESCRIPTION = "TESS-W Calibration Database Quality Assurance tool"

_CENTRAL_MAP = {
    CentralTendency.MEAN: statistics.mean,
    CentralTendency.MEDIAN: statistics.median,
    CentralTendency.MODE: statistics.mode,
}

# -----------------------
# Module global variables
# -----------------------

# get the root logger
log = logging.getLogger(__name__.split(".")[-1])

# ------------------
# Auxiliar functions
# ------------------


def central(method: CentralTendency):
    assert method in CentralTendency, f"Statistics central method {method} is not supported"
    return _CENTRAL_MAP[method]


def magnitude(zp, freq):
    return zp - 2.5 * math.log10(freq)


# -----------------
# Auxiliary classes
# -----------------


class DbgPhotometer(Photometer):
    pass


class DbgSummary(Summary):
    def assert_nrounds(self, rounds):
        N = len(rounds)
        if not (self.nrounds is None or self.nrounds == N):
            log.error(
                "[%s] [%s] [%s] Summary rounds. computed = %d, stored = %d",
                self.n,
                self.m,
                self.s,
                N,
                self.nrounds,
            )

    def assert_fict_zp(self, rounds, default: float = 20.50):
        if not all([r.zp_fict == rounds[0].zp_fict for r in rounds]):
            log.error(
                "[%s] [%s] [%s] Summary. Fict ZP: All not are equal %s",
                self.n,
                self.m,
                self.s,
                [r.zp_fict for r in rounds],
            )
        if rounds:
            zp_fict = rounds[0].zp_fict
            if rounds[0].zp_fict != default:
                log.warn(
                    "[%s] [%s] [%s] Summary. Ficticious ZP (%s) is not %s",
                    self.n,
                    self.m,
                    self.s,
                    zp_fict,
                    default,
                )

    def assert_freq_from_rounds(self, rounds):
        """
        Asserts that the computed central frequency from the several frequency rounds
        using the central method also specied in the Summary table
        equals the central frequency stored in the Summary table.
        """
        freqs = [r.freq for r in rounds]
        central_func = central(self.freq_method)
        freq = central_func(freqs)
        if not math.fabs(freq - self.freq) < 0.0005:
            log.error(
                "[%s] [%s] [%s]  Summary Frequency: computed =%.3f, stored =%.3f, \u0394 = %.3f",
                self.n,
                self.m,
                self.s,
                freq,
                self.freq,
                freq - self.freq,
            )
        return freq

    def assert_mag_from_rounds(self, rounds, freq):
        zp_fict = rounds[0].zp_fict
        mag = magnitude(zp_fict, freq)
        if not math.fabs(mag - self.mag) < 0.005:
            log.warn(
                "[%s] [%s] [%s]  Summary Magnitudes: computed =%f from f=%.3f & zp=%.2f, stored mag=%f from stored f=%.3f \u0394 = %.3f",
                self.n,
                self.m,
                self.s,
                mag,
                freq,
                zp_fict,
                self.mag,
                self.freq,
                mag - self.mag,
            )

    def assert_zp_from_rounds(self, rounds):
        """
        Asserts that the computed central zero point from the several zero point rounds
        using the central method also specied in the Summary table
        equals the central zero_point stored in the Summary table (minus the offset).
        """
        zps = [r.zero_point for r in rounds]
        central_func = central(self.zero_point_method)
        zp = central_func(zps) + self.zp_offset
        if not math.fabs(zp - self.zero_point) < 0.005:
            log.warn(
                "[%s] [%s] [%s]  Summary Zero Points: computed zp=%f, stored zp=%f",
                self.n,
                self.m,
                self.s,
                zp,
                self.zero_point,
            )

    async def check(self, photometer):
        self.n = photometer.name
        self.m = photometer.mac
        self.s = self.session
        rounds = await self.awaitable_attrs.rounds
        log.info("[%s] [%s] [%s] Summary assert_nounds", self.n, self.m, self.s)
        self.assert_nrounds(rounds)
        log.info("[%s] [%s] [%s] Summary assert_fict_zp", self.n, self.m, self.s)
        self.assert_fict_zp(rounds)
        if self.nrounds is not None:
            log.info("[%s] [%s] [%s] Summary assert_freq_from_rounds", self.n, self.m, self.s)
            freq = self.assert_freq_from_rounds(rounds)
            self.assert_mag_from_rounds(rounds, freq)
            if self.role == Role.TEST:
                log.info("[%s] [%s] [%s] Summary assert_zp_from_rounds", self.n, self.m, self.s)
                self.assert_zp_from_rounds(rounds)


class DbgRound(Round):
    def assert_round_magnitude(self) -> float:
        mag = self.zp_fict - 2.5 * math.log10(self.freq)
        if not math.fabs(self.mag - mag) < 0.005:
            log.error(
                "[%s] [%s] [%s] Round #%d. Magnitudes: computed = %f @ zp = %f, stored = %f",
                self.n,
                self.m,
                self.s,
                self.seq,
                mag,
                self.zp_fict,
                self.mag,
            )

    def assert_freq_from_samples(self, samples) -> float:
        """Computes the central frequnency from its samples"""
        freqs = [s.freq for s in samples]
        central_func = central(self.central)
        freq = central_func(freqs)
        if not math.fabs(self.freq - freq) < 0.0005:
            log.error(
                "[%s] [%s] [%s] Round #%d. %s Frequency: computed f = %f, stored = %f",
                self.n,
                self.m,
                self.s,
                self.seq,
                self.central,
                freq,
                self.freq,
            )
        stddev = statistics.stdev(freqs, freq)
        if math.fabs(self.stddev - stddev) > 0.005:
            log.warn(
                "[%s] [%s] [%s] Round #%d. computed \u03c3(freq) = %f, stored \u03c3(freq) = %f. May be can be fixed.",
                self.n,
                self.m,
                self.s,
                self.seq,
                stddev,
                self.stddev,
            )
            stddev2 = statistics.stdev(freqs)
            if not math.fabs(self.stddev - stddev2) < 0.005:
                log.error(
                    "[%s] [%s] [%s] Round #%d. computed \u03c3(freq) = %f, stored \u03c3(freq) = %f",
                    self.n,
                    self.m,
                    self.s,
                    self.seq,
                    stddev2,
                    self.stddev,
                )

    def assert_no_timestamps(self):
        if not (self.begin_tstamp is None and self.end_tstamp is None):
            log.error(
                "[%s] [%s] [%s] Round #%d. Expected empty time window, got from %s to %s",
                self.n,
                self.m,
                self.s,
                self.seq,
                self.begin_tstamp,
                self.end_tstamp,
            )

    def assert_samples(self, samples):
        N = len(samples)
        if not self.nsamples == N:
            log.error(
                "[%s] [%s] [%s] Round #%d. Number of samples. Computed = %d, Stored = %d",
                self.n,
                self.m,
                self.s,
                self.seq,
                N,
                self.nsamples,
            )
        if not self.begin_tstamp == samples[0].tstamp:
            log.error(
                "[%s] [%s] [%s] Round #%d. Begin round timestamp mismatch. Round = %s, Samples[0] = %s",
                self.n,
                self.m,
                self.s,
                self.seq,
                self.begin_tstamp,
                samples[0].tstamp,
            )
        if not self.end_tstamp == samples[-1].tstamp:
            log.error(
                "[%s] [%s] [%s] Round #%d. End round timestamp mismatch. Round = %s, Samples[-1] = %s",
                self.n,
                self.m,
                self.s,
                self.seq,
                self.end_tstamp,
                samples[-1].tstamp,
            )
        for s in samples:
            if not s.role == self.role:
                log.error(
                    "[%s] [%s] [%s] Round #%d. Wrong roles Round = %s, Samples[s] = %s",
                    self.n,
                    self.m,
                    self.s,
                    self.seq,
                    self.role,
                    s.role,
                )

    async def check(self, photometer, summary):
        self.n = photometer.name
        self.m = photometer.mac
        self.s = summary.session
        log.info("[%s] [%s] [%s] Round #%d self check", self.n, self.m, self.s, self.seq)
        self.assert_round_magnitude()
        total_samples = await summary.awaitable_attrs.samples
        if self.nsamples > 0 and len(total_samples) == 0:
            self.assert_no_timestamps()
            log.info(
                "[%s] [%s] [%s] Round #%d self check ok, but 0/%d STORED SAMPLES !",
                self.n,
                self.m,
                self.s,
                self.seq,
                self.nsamples,
            )
            return
        samples = sorted(await self.awaitable_attrs.samples)
        self.assert_samples(samples)
        self.assert_freq_from_samples(samples)


class DbgSample(Sample):
    async def check(self, photometer, summary):
        n = photometer.name
        m = photometer.mac
        s = summary.session
        rounds = await self.awaitable_attrs.rounds
        rseqs = sorted([r.seq for r in rounds])
        log.info("[%s] [%s] [%s] Sample #%d in Rounds %s. self check ok", n, m, s, self.id, rseqs)


# -------------------
# Auxiliary functions
# -------------------


async def get_all_sessions(
    async_session: async_sessionmaker[AsyncSessionClass],
) -> List[datetime]:
    async with async_session() as session:
        async with session.begin():
            q = select(DbgSummary.session.distinct()).order_by(DbgSummary.role.asc())
            return (await session.scalars(q)).all()


async def check_summary(
    meas_session: datetime, async_session: async_sessionmaker[AsyncSessionClass]
) -> None:
    if meas_session is not None:
        await check_summary_single(meas_session, async_session)
    else:
        meas_session = await get_all_sessions(async_session)
        for ses in meas_session:
            await check_summary_single(ses, async_session)


async def check_rounds(
    meas_session: datetime, async_session: async_sessionmaker[AsyncSessionClass]
) -> None:
    if meas_session is not None:
        await check_rounds_single(meas_session, async_session)
    else:
        meas_session = await get_all_sessions(async_session)
        for ses in meas_session:
            await check_rounds_single(ses, async_session)


async def check_samples(
    meas_session: datetime, async_session: async_sessionmaker[AsyncSessionClass]
) -> None:
    if meas_session is not None:
        await check_samples_single(meas_session, async_session)
    else:
        meas_session = await get_all_sessions(async_session)
        for ses in meas_session:
            await check_samples_single(ses, async_session)


async def check_all(
    meas_session: datetime, async_session: async_sessionmaker[AsyncSessionClass]
) -> None:
    if meas_session is not None:
        await check_all_single(meas_session, async_session)
    else:
        meas_session = await get_all_sessions(async_session)
        for ses in meas_session:
            await check_all_single(ses, async_session)


async def check_summary_single(
    meas_session: datetime, async_session: async_sessionmaker[AsyncSessionClass]
) -> None:
    async with async_session() as session:
        async with session.begin():
            q = (
                select(DbgPhotometer, DbgSummary)
                .join(DbgSummary)
                .where(DbgSummary.session == meas_session)
                .order_by(DbgSummary.role.asc())
            )
            result = (await session.execute(q)).all()
            for row in result:
                photometer = row[0]
                summary = row[1]
                await summary.check(photometer)


async def check_rounds_single(
    meas_session: datetime, async_session: async_sessionmaker[AsyncSessionClass]
) -> None:
    async with async_session() as session:
        async with session.begin():
            q = (
                select(DbgPhotometer, DbgSummary, DbgRound)
                .join(DbgSummary, DbgPhotometer.id == DbgSummary.phot_id)
                .join(DbgRound, DbgSummary.id == DbgRound.summ_id)
                .filter(DbgSummary.session == meas_session)
                .order_by(DbgSummary.role.asc())
            )
            result = (await session.execute(q)).all()
            for row in result:
                photometer = row[0]
                summary = row[-2]
                round_ = row[-1]
                await round_.check(photometer, summary)


async def check_samples_single(
    meas_session: datetime, async_session: async_sessionmaker[AsyncSessionClass]
) -> None:
    async with async_session() as session:
        async with session.begin():
            q = (
                select(DbgPhotometer, DbgSummary, DbgSample)
                .join(DbgSummary, DbgPhotometer.id == DbgSummary.phot_id)
                .join(DbgSample, DbgSummary.id == DbgSample.summ_id)
                .filter(DbgSummary.session == meas_session)
                .order_by(DbgSummary.role.asc())
            )
            result = (await session.execute(q)).all()
            for row in result:
                photometer = row[0]
                summary = row[-2]
                sample = row[-1]
                await sample.check(photometer, summary)


async def check_all_single(
    meas_session: datetime, async_session: async_sessionmaker[AsyncSessionClass]
) -> None:
    await check_summary_single(meas_session, async_session)
    await check_rounds_single(meas_session, async_session)
    await check_samples_single(meas_session, async_session)


# --------------
# main functions
# --------------

TABLE = {
    "summary": check_summary,
    "rounds": check_rounds,
    "samples": check_samples,
}


async def qa(args) -> None:
    if args.command != "all":
        func = TABLE[args.command]
        meas_session = args.session
        await func(meas_session, AsyncSession)
    else:
        for name in ("summary", "rounds", "samples"):
            meas_session = args.session
            func = TABLE[name]
            await func(meas_session, AsyncSession)


def add_args(parser):
    subparser = parser.add_subparsers(dest="command")
    subparser.add_parser("summary", parents=[prs_session()], help="Browse summary data")
    subparser.add_parser("rounds", parents=[prs_session()], help="Browse rounds data")
    subparser.add_parser("samples", parents=[prs_session()], help="Browse samples data")
    subparser.add_parser("all", parents=[prs_session()], help="Browse all data")


def prs_session() -> ArgumentParser:
    parser = ArgumentParser(add_help=False)
    parser.add_argument(
        "-s",
        "--session",
        metavar="<YYYY-MM-DDTHH:MM:SS>",
        type=vdate,
        default=None,
        help="Session date",
    )
    return parser


async def cli_main(args: Namespace) -> None:
    sqa_logging(args)
    await qa(args)


def main():
    """The main entry point specified by pyproject.toml"""
    execute(
        main_func=cli_main,
        add_args_func=add_args,
        name=__name__,
        version=__version__,
        description=DESCRIPTION,
    )


if __name__ == "__main__":
    main()
